@@ -2,11 +2,22 @@ import asyncio
 import calendar
 import collections
 import contextlib
-import datetime
 import logging
-import pathlib
 import re
 import ssl
+import sys
+from collections.abc import AsyncGenerator, Sequence
+from datetime import datetime
+from pathlib import Path, PurePosixPath
+from types import TracebackType
+from typing import (
+    Any,
+    Literal,
+    Protocol,
+    TypedDict,
+    Union,
+    overload,
+)
 
 from . import errors, pathio
 from .common import (
@@ -18,15 +29,23 @@ from .common import (
     END_OF_LINE,
     HALF_OF_YEAR_IN_SECONDS,
     TWO_YEARS_IN_SECONDS,
-    AsyncListerMixin,
+    AbstractAsyncLister,
+    AsyncEnterableInstanceProtocol,
+    Code,
     SSLSessionBoundContext,
     StreamThrottle,
     ThrottleStreamIO,
     async_enterable,
     setlocale,
+    wrap_into_codes,
     wrap_with_container,
 )
 
+if sys.version_info >= (3, 11):
+    from typing import NotRequired, Self, Unpack
+else:
+    from typing_extensions import NotRequired, Self, Unpack
+
 try:
     from siosocks.io.asyncio import open_connection
 except ImportError:
@@ -37,32 +56,10 @@ __all__ = (
     "BaseClient",
     "Client",
     "DataConnectionThrottleStreamIO",
-    "Code",
 )
 logger = logging.getLogger(__name__)
 
 
-class Code(str):
-    """
-    Representation of server status code.
-    """
-
-    def matches(self, mask):
-        """
-        :param mask: Template for comparision. If mask symbol is not digit
-            then it passes.
-        :type mask: :py:class:`str`
-
-        ::
-
-            >>> Code("123").matches("1")
-            True
-            >>> Code("123").matches("1x3")
-            True
-        """
-        return all(map(lambda m, c: not m.isdigit() or m == c, mask, self))
-
-
 class DataConnectionThrottleStreamIO(ThrottleStreamIO):
     """
     Add `finish` method to :py:class:`aioftp.ThrottleStreamIO`, which is
@@ -78,11 +75,32 @@ class DataConnectionThrottleStreamIO(ThrottleStreamIO):
         :py:class:`aioftp.ThrottleStreamIO`
     """
 
-    def __init__(self, client, *args, **kwargs):
-        super().__init__(*args, **kwargs)
+    def __init__(
+        self,
+        client: "BaseClient",
+        reader: asyncio.StreamReader,
+        writer: asyncio.StreamWriter,
+        throttles: dict[str, StreamThrottle],
+        *,
+        timeout: Union[float, int, None] = None,
+        read_timeout: Union[float, int, None] = None,
+        write_timeout: Union[float, int, None] = None,
+    ) -> None:
+        super().__init__(
+            reader,
+            writer,
+            throttles,
+            timeout=timeout,
+            read_timeout=read_timeout,
+            write_timeout=write_timeout,
+        )
         self.client = client
 
-    async def finish(self, expected_codes="2xx", wait_codes="1xx"):
+    async def finish(
+        self,
+        expected_codes: Union[str, tuple[str, ...]] = "2xx",
+        wait_codes: Union[str, tuple[str, ...]] = "1xx",
+    ) -> None:
         """
         :py:func:`asyncio.coroutine`
 
@@ -100,29 +118,88 @@ class DataConnectionThrottleStreamIO(ThrottleStreamIO):
         self.close()
         await self.client.command(None, expected_codes, wait_codes)
 
-    async def __aexit__(self, exc_type, exc, tb):
+    async def __aexit__(
+        self,
+        exc_type: Union[type[BaseException], None],
+        exc: Union[BaseException, None],
+        tb: TracebackType,
+    ) -> None:
         if exc is None:
             await self.finish()
         else:
             self.close()
 
 
+UnixListInfo = TypedDict(
+    "UnixListInfo",
+    {
+        "type": str,
+        "unix.mode": int,
+        "unix.links": str,
+        "unix.owner": str,
+        "unix.group": str,
+        "size": str,
+        "modify": str,
+        "link_dst": NotRequired[str],
+    },
+)
+
+
+class BasicListInfo(TypedDict):
+    modify: str
+    type: str
+    create: NotRequired[str]
+    size: str
+
+
+ListInfo = Union[BasicListInfo, UnixListInfo]
+
+
+PathWithBasicInfo = tuple[PurePosixPath, BasicListInfo]
+
+
+class ParseListLineCustomCallable(Protocol):
+    def __call__(self, b: bytes) -> PathWithBasicInfo: ...
+
+
+PathWithInfo = tuple[PurePosixPath, Union[BasicListInfo, UnixListInfo]]
+
+
+class ParseListLineCallable(Protocol):
+    def __call__(self, b: bytes) -> PathWithInfo: ...
+
+
+class BaseClientArgs(TypedDict, total=False):
+    socket_timeout: Union[float, int, None]
+    connection_timeout: Union[float, int, None]
+    read_speed_limit: Union[int, None]
+    write_speed_limit: Union[int, None]
+    path_timeout: Union[float, int, None]
+    path_io_factory: type[pathio.AbstractPathIO[Path]]
+    encoding: str
+    ssl: Union[ssl.SSLContext, bool, None]
+    parse_list_line_custom: Union[ParseListLineCustomCallable, None]
+    parse_list_line_custom_first: bool
+    passive_commands: tuple[str, ...]
+    siosocks_asyncio_kwargs: Any
+
+
 class BaseClient:
     def __init__(
         self,
         *,
-        socket_timeout=None,
-        connection_timeout=None,
-        read_speed_limit=None,
-        write_speed_limit=None,
-        path_timeout=None,
-        path_io_factory=pathio.PathIO,
-        encoding="utf-8",
-        ssl=None,
-        parse_list_line_custom=None,
-        parse_list_line_custom_first=True,
-        passive_commands=("epsv", "pasv"),
-        **siosocks_asyncio_kwargs,
+        socket_timeout: Union[float, int, None] = None,
+        connection_timeout: Union[float, int, None] = None,
+        read_speed_limit: Union[int, None] = None,
+        write_speed_limit: Union[int, None] = None,
+        path_timeout: Union[float, int, None] = None,
+        path_io_factory: type[pathio.AbstractPathIO[Path]] = pathio.PathIO,
+        encoding: str = "utf-8",
+        ssl: Union[ssl.SSLContext, bool, None] = None,
+        parse_list_line_custom: Union[ParseListLineCustomCallable, None] = None,
+        parse_list_line_custom_first: bool = True,
+        passive_commands: tuple[str, ...] = ("epsv", "pasv"),
+        **siosocks_asyncio_kwargs: Any,
     ):
         self.socket_timeout = socket_timeout
         self.connection_timeout = connection_timeout
@@ -133,48 +210,59 @@ class BaseClient:
         self.path_timeout = path_timeout
         self.path_io = path_io_factory(timeout=path_timeout)
         self.encoding = encoding
-        self.stream = None
+        self._stream: Union[ThrottleStreamIO, None] = None
         self.ssl = ssl
         self.parse_list_line_custom = parse_list_line_custom
         self.parse_list_line_custom_first = parse_list_line_custom_first
         self._passive_commands = passive_commands
         self._siosocks_asyncio_kwargs = siosocks_asyncio_kwargs
 
-    async def _open_connection(self, host, port):
+    async def _open_connection(self, host: str, port: int) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
         ssl_resolved = self.ssl
-        if self.stream is not None:
-            ssl_object = self.stream.writer.transport.get_extra_info("ssl_object")
+        if self._stream is not None:
+            ssl_object = self._stream.writer.transport.get_extra_info("ssl_object")
             if ssl_object is not None:
                 ssl_resolved = SSLSessionBoundContext(
                     ssl.PROTOCOL_TLS_CLIENT,
                     context=ssl_object.context,
                     session=ssl_object.session,
                 )
-        connection = await open_connection(host, port, ssl=ssl_resolved, **self._siosocks_asyncio_kwargs)
+        connection: tuple[asyncio.StreamReader, asyncio.StreamWriter] = await open_connection(
+            host,
+            port,
+            ssl=ssl_resolved,
+            **self._siosocks_asyncio_kwargs,
+        )
         return connection
 
-    async def connect(self, host, port=DEFAULT_PORT):
+    async def connect(self, host: str, port: int = DEFAULT_PORT) -> None:
         self.server_host = host
         self.server_port = port
         reader, writer = await asyncio.wait_for(
             self._open_connection(host, port),
             self.connection_timeout,
         )
-        self.stream = ThrottleStreamIO(
+        self._stream = ThrottleStreamIO(
             reader,
             writer,
             throttles={"_": self.throttle},
             timeout=self.socket_timeout,
         )
 
-    def close(self):
+    @property
+    def stream(self) -> ThrottleStreamIO:
+        if self._stream is None:
+            raise ConnectionError("Connect first")
+        return self._stream
+
+    def close(self) -> None:
         """
         Close connection.
         """
-        if self.stream is not None:
-            self.stream.close()
+        if self._stream is not None:
+            self._stream.close()
 
-    async def parse_line(self):
+    async def parse_line(self) -> tuple[Code, str]:
         """
         :py:func:`asyncio.coroutine`
 
@@ -196,7 +284,7 @@ class BaseClient:
         logger.debug(s)
         return Code(s[:3]), s[3:]
 
-    async def parse_response(self):
+    async def parse_response(self) -> tuple[Code, list[str]]:
         """
         :py:func:`asyncio.coroutine`
 
@@ -221,7 +309,7 @@ class BaseClient:
                 info.append(curr_code + rest)
         return code, info
 
-    def check_codes(self, expected_codes, received_code, info):
+    def check_codes(self, expected_codes: Union[Code, tuple[Code, ...]], received_code: Code, info: list[str]) -> None:
         """
         Checks if any of expected matches received.
 
@@ -240,13 +328,31 @@ class BaseClient:
         if not any(map(received_code.matches, expected_codes)):
             raise errors.StatusCodeError(expected_codes, received_code, info)
 
+    @overload
     async def command(
         self,
-        command=None,
-        expected_codes=(),
-        wait_codes=(),
-        censor_after=None,
-    ):
+        command: Union[str, None] = None,
+        expected_codes: tuple[()] = (),
+        wait_codes: tuple[()] = (),
+        censor_after: None = None,
+    ) -> None: ...
+
+    @overload
+    async def command(
+        self,
+        command: Union[str, None] = None,
+        expected_codes: Union[str, tuple[str, ...]] = (),
+        wait_codes: Union[str, tuple[str, ...]] = (),
+        censor_after: Union[int, None] = None,
+    ) -> tuple[Code, list[str]]: ...
+
+    async def command(
+        self,
+        command: Union[str, None] = None,
+        expected_codes: Union[str, tuple[str, ...]] = (),
+        wait_codes: Union[str, tuple[str, ...]] = (),
+        censor_after: Union[int, None] = None,
+    ) -> Union[tuple[Code, list[str]], None]:
         """
         :py:func:`asyncio.coroutine`
 
@@ -271,8 +377,8 @@ class BaseClient:
             when logging
         :type censor_after: :py:class:`None` or :py:class:`int`
         """
-        expected_codes = wrap_with_container(expected_codes)
-        wait_codes = wrap_with_container(wait_codes)
+        expected_codes = wrap_into_codes(wrap_with_container(expected_codes))
+        wait_codes = wrap_into_codes(wrap_with_container(wait_codes))
         if command:
             if censor_after:
                 # Censor the user's command
@@ -290,9 +396,10 @@ class BaseClient:
             if expected_codes:
                 self.check_codes(expected_codes, code, info)
             return code, info
+        return None
 
     @staticmethod
-    def parse_epsv_response(s):
+    def parse_epsv_response(s: str) -> tuple[None, int]:
         """
         Parsing `EPSV` (`message (|||port|)`) response.
 
@@ -308,7 +415,7 @@ class BaseClient:
         return None, port
 
     @staticmethod
-    def parse_pasv_response(s):
+    def parse_pasv_response(s: str) -> tuple[str, int]:
         """
         Parsing `PASV` server response.
 
@@ -325,7 +432,7 @@ class BaseClient:
         return ip, port
 
     @staticmethod
-    def parse_directory_response(s):
+    def parse_directory_response(s: str) -> PurePosixPath:
         """
         Parsing directory server response.
 
@@ -351,12 +458,12 @@ class BaseClient:
                         seq_quotes = 0
                         directory += '"'
                     directory += ch
-        return pathlib.PurePosixPath(directory)
+        return PurePosixPath(directory)
 
     @staticmethod
-    def parse_unix_mode(s):
+    def parse_unix_mode(s: str) -> int:
         """
-        Parsing unix mode strings ("rwxr-x--t") into hexacimal notation.
+        Parsing unix mode strings ("rwxr-x--t") into hexadecimal notation.
 
         :param s: mode string
         :type s: :py:class:`str`
@@ -393,7 +500,7 @@ class BaseClient:
         return mode
 
     @staticmethod
-    def format_date_time(d):
+    def format_date_time(d: datetime) -> str:
         """
         Formats dates from strptime in a consistent format
 
@@ -405,7 +512,7 @@ class BaseClient:
         return d.strftime("%Y%m%d%H%M00")
 
     @classmethod
-    def parse_ls_date(cls, s, *, now=None):
+    def parse_ls_date(cls, s: str, *, now: Union[datetime, None] = None) -> str:
         """
         Parsing dates from the ls unix utility. For example,
         "Nov 18  1958", "Jan 03 2018", and "Nov 18 12:29".
@@ -418,13 +525,13 @@ class BaseClient:
         with setlocale("C"):
             try:
                 if now is None:
-                    now = datetime.datetime.now()
+                    now = datetime.now()
                 if s.startswith("Feb 29"):
                     # Need to find the nearest previous leap year
                     prev_leap_year = now.year
                     while not calendar.isleap(prev_leap_year):
                         prev_leap_year -= 1
-                    d = datetime.datetime.strptime(
+                    d = datetime.strptime(
                         f"{prev_leap_year} {s}",
                         "%Y %b %d %H:%M",
                     )
@@ -433,22 +540,22 @@ class BaseClient:
                     if diff > TWO_YEARS_IN_SECONDS:
                         d = d.replace(year=prev_leap_year + 4)
                 else:
-                    d = datetime.datetime.strptime(f"{now.year} {s}", "%Y %b %d %H:%M")
+                    d = datetime.strptime(f"{now.year} {s}", "%Y %b %d %H:%M")
                     diff = (now - d).total_seconds()
                     if diff > HALF_OF_YEAR_IN_SECONDS:
                         d = d.replace(year=now.year + 1)
                     elif diff < -HALF_OF_YEAR_IN_SECONDS:
                         d = d.replace(year=now.year - 1)
             except ValueError:
-                d = datetime.datetime.strptime(s, "%b %d  %Y")
+                d = datetime.strptime(s, "%b %d  %Y")
         return cls.format_date_time(d)
 
-    def parse_list_line_unix(self, b):
+    def parse_list_line_unix(self, b: bytes) -> tuple[PurePosixPath, UnixListInfo]:
         """
         Attempt to parse a LIST line (similar to unix ls utility).
 
         :param b: response line
-        :type b: :py:class:`bytes` or :py:class:`str`
+        :type b: :py:class:`bytes`
 
         :return: (path, info)
         :rtype: (:py:class:`pathlib.PurePosixPath`, :py:class:`dict`)
@@ -465,7 +572,7 @@ class BaseClient:
             info["type"] = "unknown"
 
         # TODO: handle symlinks(beware the symlink loop)
-        info["unix.mode"] = self.parse_unix_mode(s[1:10])
+        info["unix.mode"] = self.parse_unix_mode(s[1:10])  # type: ignore[assignment]
         s = s[10:].lstrip()
         i = s.index(" ")
         info["unix.links"] = s[:i]
@@ -497,25 +604,25 @@ class BaseClient:
             info["type"] = "dir" if link_dst[i] == "/" else "file"
             info["link_dst"] = link_dst
             s = link_src
-        return pathlib.PurePosixPath(s), info
+        return PurePosixPath(s), info  # type: ignore[return-value]
 
-    def parse_list_line_windows(self, b):
+    def parse_list_line_windows(self, b: bytes) -> tuple[PurePosixPath, BasicListInfo]:
         """
         Parsing Microsoft Windows `dir` output
 
         :param b: response line
-        :type b: :py:class:`bytes` or :py:class:`str`
+        :type b: :py:class:`bytes`
 
         :return: (path, info)
         :rtype: (:py:class:`pathlib.PurePosixPath`, :py:class:`dict`)
         """
         line = b.decode(encoding=self.encoding).rstrip("\r\n")
+
         date_time_end = line.index("M")
-        date_time_str = line[: date_time_end + 1].strip().split(" ")
-        date_time_str = " ".join([x for x in date_time_str if len(x) > 0])
+        date_time_str = " ".join(line[: date_time_end + 1].strip().split(" "))
         line = line[date_time_end + 1 :].lstrip()
         with setlocale("C"):
-            strptime = datetime.datetime.strptime
+            strptime = datetime.strptime
             date_time = strptime(date_time_str, "%m/%d/%Y %I:%M %p")
         info = {}
         info["modify"] = self.format_date_time(date_time)
@@ -533,28 +640,29 @@ class BaseClient:
         filename = line[next_space:].lstrip()
         if filename == "." or filename == "..":
             raise ValueError
-        return pathlib.PurePosixPath(filename), info
+        return PurePosixPath(filename), info  # type: ignore[return-value]
 
-    def parse_list_line(self, b):
+    def parse_list_line(self, b: bytes) -> PathWithInfo:
         """
         Parse LIST response with both Microsoft Windows® parser and
         UNIX parser
 
         :param b: response line
-        :type b: :py:class:`bytes` or :py:class:`str`
+        :type b: :py:class:`bytes`
 
         :return: (path, info)
         :rtype: (:py:class:`pathlib.PurePosixPath`, :py:class:`dict`)
         """
         ex = []
-        parsers = [
+        parsers: list[Union[ParseListLineCallable, None]] = [
             self.parse_list_line_unix,
             self.parse_list_line_windows,
         ]
-        if self.parse_list_line_custom_first:
-            parsers = [self.parse_list_line_custom] + parsers
-        else:
-            parsers = parsers + [self.parse_list_line_custom]
+        if self.parse_list_line_custom is not None:
+            if self.parse_list_line_custom_first:
+                parsers = [self.parse_list_line_custom] + parsers
+            else:
+                parsers = parsers + [self.parse_list_line_custom]
         for parser in parsers:
             if parser is None:
                 continue
@@ -564,7 +672,7 @@ class BaseClient:
                 ex.append(e)
         raise ValueError("All parsers failed to parse", b, ex)
 
-    def parse_mlsx_line(self, b):
+    def parse_mlsx_line(self, b: Union[bytes, str]) -> tuple[PurePosixPath, BasicListInfo]:
         """
         Parsing MLS(T|D) response.
 
@@ -584,7 +692,13 @@ class BaseClient:
         for fact in facts_found[:-1].split(";"):
             key, _, value = fact.partition("=")
             entry[key.lower()] = value
-        return pathlib.PurePosixPath(name), entry
+        return PurePosixPath(name), entry  # type: ignore[return-value]
+
+
+ConnType = Union[Literal["I"], Literal["A"], Literal["E"], Literal["L"]]
+
+
+PathLike = Union[str, PurePosixPath]
 
 
 class Client(BaseClient):
@@ -623,7 +737,7 @@ class Client(BaseClient):
     :type ssl: :py:class:`bool` or :py:class:`ssl.SSLContext`
 
     :param parse_list_line_custom: callable, which receive exactly one
-        argument: line of type bytes. Should return tuple of Path object and
+        argument: line of type bytes or str. Should return tuple of Path object and
         dictionary with fields "modify", "type", "size". For more
         information see sources.
     :type parse_list_line_custom: callable
@@ -633,7 +747,7 @@ class Client(BaseClient):
     :param **siosocks_asyncio_kwargs: siosocks key-word only arguments
     """
 
-    async def connect(self, host, port=DEFAULT_PORT):
+    async def connect(self, host: str, port: int = DEFAULT_PORT) -> list[str]:  # type: ignore[override]
         """
         :py:func:`asyncio.coroutine`
 
@@ -658,7 +772,7 @@ class Client(BaseClient):
         await self.command("PBSZ 0", "200")
         await self.command("PROT P", "200")
 
-    async def upgrade_to_tls(self, sslcontext=None):
+    async def upgrade_to_tls(self, sslcontext: Union[ssl.SSLContext, None] = None) -> None:
         """
         :py:func:`asyncio.coroutine`
 
@@ -682,10 +796,10 @@ class Client(BaseClient):
 
     async def login(
         self,
-        user=DEFAULT_USER,
-        password=DEFAULT_PASSWORD,
-        account=DEFAULT_ACCOUNT,
-    ):
+        user: str = DEFAULT_USER,
+        password: str = DEFAULT_PASSWORD,
+        account: str = DEFAULT_ACCOUNT,
+    ) -> None:
         """
         :py:func:`asyncio.coroutine`
 
@@ -711,14 +825,14 @@ class Client(BaseClient):
             elif code == "332":
                 cmd = "ACCT " + account
             else:
-                raise errors.StatusCodeError("33x", code, info)
+                raise errors.StatusCodeError(Code("33x"), code, info)
             code, info = await self.command(
                 cmd,
                 ("230", "33x"),
                 censor_after=censor_after,
             )
 
-    async def get_current_directory(self):
+    async def get_current_directory(self) -> PurePosixPath:
         """
         :py:func:`asyncio.coroutine`
 
@@ -730,7 +844,7 @@ class Client(BaseClient):
         directory = self.parse_directory_response(info[-1])
         return directory
 
-    async def change_directory(self, path=".."):
+    async def change_directory(self, path: PathLike = "..") -> None:
         """
         :py:func:`asyncio.coroutine`
 
@@ -739,14 +853,14 @@ class Client(BaseClient):
         :param path: new directory, goes «up» if omitted
         :type path: :py:class:`str` or :py:class:`pathlib.PurePosixPath`
         """
-        path = pathlib.PurePosixPath(path)
-        if path == pathlib.PurePosixPath(".."):
+        path = PurePosixPath(path)
+        if path == PurePosixPath(".."):
             cmd = "CDUP"
         else:
             cmd = "CWD " + str(path)
         await self.command(cmd, "2xx")
 
-    async def make_directory(self, path, *, parents=True):
+    async def make_directory(self, path: PathLike, *, parents: bool = True) -> None:
         """
         :py:func:`asyncio.coroutine`
 
@@ -758,7 +872,7 @@ class Client(BaseClient):
         :param parents: create parents if does not exists
         :type parents: :py:class:`bool`
         """
-        path = pathlib.PurePosixPath(path)
+        path = PurePosixPath(path)
         need_create = []
         while path.name and not await self.exists(path):
             need_create.append(path)
@@ -769,7 +883,7 @@ class Client(BaseClient):
         for path in need_create:
             await self.command("MKD " + str(path), "257")
 
-    async def remove_directory(self, path):
+    async def remove_directory(self, path: PathLike) -> None:
         """
         :py:func:`asyncio.coroutine`
 
@@ -780,7 +894,13 @@ class Client(BaseClient):
         """
         await self.command("RMD " + str(path), "250")
 
-    def list(self, path="", *, recursive=False, raw_command=None):
+    def list(
+        self,
+        path: PathLike = "",
+        *,
+        recursive: bool = False,
+        raw_command: Union[Literal["MLSD"], Literal["LIST"], None] = None,
+    ) -> AbstractAsyncLister[PathWithInfo]:
         """
         :py:func:`asyncio.coroutine`
 
@@ -815,12 +935,12 @@ class Client(BaseClient):
             >>> stats = await client.list()
         """
 
-        class AsyncLister(AsyncListerMixin):
-            stream = None
+        class AsyncLister(AbstractAsyncLister[PathWithInfo]):
+            stream: Union[DataConnectionThrottleStreamIO, None] = None
 
-            async def _new_stream(cls, local_path):
+            async def _new_stream(cls, local_path: PathLike) -> DataConnectionThrottleStreamIO:  # type: ignore[return]
                 cls.path = local_path
-                cls.parse_line = self.parse_mlsx_line
+                cls.parse_line: ParseListLineCallable = self.parse_mlsx_line
                 if raw_command not in [None, "MLSD", "LIST"]:
                     raise ValueError(
                         f"raw_command must be one of MLSD or LIST, but got {raw_command}",
@@ -838,11 +958,13 @@ class Client(BaseClient):
                     command = ("LIST " + str(cls.path)).strip()
                     return await self.get_stream(command, "1xx")
 
-            def __aiter__(cls):
-                cls.directories = collections.deque()
+            def __aiter__(cls) -> Self:
+                cls.directories: collections.deque[tuple[PurePosixPath, Union[BasicListInfo, UnixListInfo]]] = (
+                    collections.deque()
+                )
                 return cls
 
-            async def __anext__(cls):
+            async def __anext__(cls) -> PathWithInfo:
                 if cls.stream is None:
                     cls.stream = await cls._new_stream(path)
                 while True:
@@ -867,7 +989,7 @@ class Client(BaseClient):
 
         return AsyncLister()
 
-    async def stat(self, path):
+    async def stat(self, path: PathLike) -> Union[BasicListInfo, UnixListInfo]:
         """
         :py:func:`asyncio.coroutine`
 
@@ -879,18 +1001,19 @@ class Client(BaseClient):
         :return: path info
         :rtype: :py:class:`dict`
         """
-        path = pathlib.PurePosixPath(path)
+        path = PurePosixPath(path)
+        file_info: Union[BasicListInfo, UnixListInfo]
         try:
             code, info = await self.command("MLST " + str(path), "2xx")
-            name, info = self.parse_mlsx_line(info[1].lstrip())
-            return info
+            name, file_info = self.parse_mlsx_line(info[1].lstrip())
+            return file_info
         except errors.StatusCodeError as e:
             if not e.received_codes[-1].matches("50x"):
                 raise
 
-        for p, info in await self.list(path.parent):
+        for p, file_info in await self.list(path.parent):
             if p.name == path.name:
-                return info
+                return file_info
         else:
             raise errors.StatusCodeError(
                 Code("2xx"),
@@ -898,7 +1021,7 @@ class Client(BaseClient):
                 "path does not exists",
             )
 
-    async def is_file(self, path):
+    async def is_file(self, path: PathLike) -> bool:
         """
         :py:func:`asyncio.coroutine`
 
@@ -912,7 +1035,7 @@ class Client(BaseClient):
         info = await self.stat(path)
         return info["type"] == "file"
 
-    async def is_dir(self, path):
+    async def is_dir(self, path: PathLike) -> bool:
         """
         :py:func:`asyncio.coroutine`
 
@@ -926,7 +1049,7 @@ class Client(BaseClient):
         info = await self.stat(path)
         return info["type"] == "dir"
 
-    async def exists(self, path):
+    async def exists(self, path: PathLike) -> bool:
         """
         :py:func:`asyncio.coroutine`
 
@@ -945,7 +1068,11 @@ class Client(BaseClient):
                 return False
             raise
 
-    async def rename(self, source, destination):
+    async def rename(
+        self,
+        source: PathLike,
+        destination: PathLike,
+    ) -> None:
         """
         :py:func:`asyncio.coroutine`
 
@@ -960,7 +1087,7 @@ class Client(BaseClient):
         await self.command("RNFR " + str(source), "350")
         await self.command("RNTO " + str(destination), "2xx")
 
-    async def remove_file(self, path):
+    async def remove_file(self, path: PathLike) -> None:
         """
         :py:func:`asyncio.coroutine`
 
@@ -971,7 +1098,7 @@ class Client(BaseClient):
         """
         await self.command("DELE " + str(path), "2xx")
 
-    async def remove(self, path):
+    async def remove(self, path: PathLike) -> None:
         """
         :py:func:`asyncio.coroutine`
 
@@ -991,7 +1118,12 @@ class Client(BaseClient):
                         await self.remove(name)
                 await self.remove_directory(path)
 
-    def upload_stream(self, destination, *, offset=0):
+    def upload_stream(
+        self,
+        destination: PathLike,
+        *,
+        offset: int = 0,
+    ) -> AsyncEnterableInstanceProtocol[DataConnectionThrottleStreamIO]:
         """
         Create stream for write data to `destination` file.
 
@@ -1009,7 +1141,12 @@ class Client(BaseClient):
             offset=offset,
         )
 
-    def append_stream(self, destination, *, offset=0):
+    def append_stream(
+        self,
+        destination: PathLike,
+        *,
+        offset: int = 0,
+    ) -> AsyncEnterableInstanceProtocol[DataConnectionThrottleStreamIO]:
         """
         Create stream for append (write) data to `destination` file.
 
@@ -1027,7 +1164,14 @@ class Client(BaseClient):
             offset=offset,
         )
 
-    async def upload(self, source, destination="", *, write_into=False, block_size=DEFAULT_BLOCK_SIZE):
+    async def upload(
+        self,
+        source: Union[str, Path],
+        destination: PathLike = "",
+        *,
+        write_into: bool = False,
+        block_size: int = DEFAULT_BLOCK_SIZE,
+    ) -> None:
         """
         :py:func:`asyncio.coroutine`
 
@@ -1048,8 +1192,8 @@ class Client(BaseClient):
         :param block_size: block size for transaction
         :type block_size: :py:class:`int`
         """
-        source = pathlib.Path(source)
-        destination = pathlib.PurePosixPath(destination)
+        source = Path(source)
+        destination = PurePosixPath(destination)
         if not write_into:
             destination = destination / source.name
         if await self.path_io.is_file(source):
@@ -1064,9 +1208,11 @@ class Client(BaseClient):
                 src = sources.popleft()
                 async for path in self.path_io.list(src):
                     if write_into:
-                        relative = destination.name / path.relative_to(source)
+                        relative: PurePosixPath = PurePosixPath(
+                            destination.name / path.relative_to(source),
+                        )
                     else:
-                        relative = path.relative_to(source.parent)
+                        relative = PurePosixPath(path.relative_to(source.parent))
                     if await self.path_io.is_dir(path):
                         await self.make_directory(relative)
                         sources.append(path)
@@ -1078,7 +1224,12 @@ class Client(BaseClient):
                             block_size=block_size,
                         )
 
-    def download_stream(self, source, *, offset=0):
+    def download_stream(
+        self,
+        source: PathLike,
+        *,
+        offset: int = 0,
+    ) -> DataConnectionThrottleStreamIO:
         """
         :py:func:`asyncio.coroutine`
 
@@ -1092,9 +1243,16 @@ class Client(BaseClient):
 
         :rtype: :py:class:`aioftp.DataConnectionThrottleStreamIO`
         """
-        return self.get_stream("RETR " + str(source), "1xx", offset=offset)
+        return self.get_stream("RETR " + str(source), "1xx", offset=offset)  # type: ignore
 
-    async def download(self, source, destination="", *, write_into=False, block_size=DEFAULT_BLOCK_SIZE):
+    async def download(
+        self,
+        source: PathLike,
+        destination: PathLike = "",
+        *,
+        write_into: bool = False,
+        block_size: int = DEFAULT_BLOCK_SIZE,
+    ) -> None:
         """
         :py:func:`asyncio.coroutine`
 
@@ -1115,23 +1273,26 @@ class Client(BaseClient):
         :param block_size: block size for transaction
         :type block_size: :py:class:`int`
         """
-        source = pathlib.PurePosixPath(source)
-        destination = pathlib.Path(destination)
+        source = PurePosixPath(source)
+        destination_path = Path(destination)
         if not write_into:
-            destination = destination / source.name
+            destination_path = destination_path / source.name
         if await self.is_file(source):
             await self.path_io.mkdir(
-                destination.parent,
+                destination_path.parent,
                 parents=True,
                 exist_ok=True,
             )
-            async with self.path_io.open(destination, mode="wb") as file_out, self.download_stream(source) as stream:
+            async with (
+                self.path_io.open(destination_path, mode="wb") as file_out,
+                self.download_stream(source) as stream,
+            ):
                 async for block in stream.iter_by_block(block_size):
                     await file_out.write(block)
         elif await self.is_dir(source):
-            await self.path_io.mkdir(destination, parents=True, exist_ok=True)
+            await self.path_io.mkdir(destination_path, parents=True, exist_ok=True)
             for name, info in await self.list(source):
-                full = destination / name.relative_to(source)
+                full = PurePosixPath(destination_path / name.relative_to(source))
                 if info["type"] in ("file", "dir"):
                     await self.download(
                         name,
@@ -1140,7 +1301,7 @@ class Client(BaseClient):
                         block_size=block_size,
                     )
 
-    async def quit(self):
+    async def quit(self) -> None:
         """
         :py:func:`asyncio.coroutine`
 
@@ -1149,21 +1310,21 @@ class Client(BaseClient):
         await self.command("QUIT", "2xx")
         self.close()
 
-    async def _do_epsv(self):
+    async def _do_epsv(self) -> tuple[None, int]:
         code, info = await self.command("EPSV", "229")
         ip, port = self.parse_epsv_response(info[-1])
         return ip, port
 
-    async def _do_pasv(self):
+    async def _do_pasv(self) -> tuple[str, int]:
         code, info = await self.command("PASV", "227")
         ip, port = self.parse_pasv_response(info[-1])
         return ip, port
 
     async def get_passive_connection(
         self,
-        conn_type="I",
-        commands=None,
-    ):
+        conn_type: ConnType = "I",
+        commands: Union[Sequence[str], None] = None,
+    ) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
         """
         :py:func:`asyncio.coroutine`
 
@@ -1189,6 +1350,7 @@ class Client(BaseClient):
         if not commands:
             raise ValueError("No passive commands provided")
         await self.command("TYPE " + conn_type, "200")
+        ip: Union[str, None]
         for i, name in enumerate(commands, start=1):
             name = name.lower()
             if name not in functions:
@@ -1200,13 +1362,21 @@ class Client(BaseClient):
                 is_last = i == len(commands)
                 if is_last or not e.received_codes[-1].matches("50x"):
                     raise
-        if ip in ("0.0.0.0", None):
+        if ip is None or ip == "0.0.0.0":
             ip = self.server_host
         reader, writer = await self._open_connection(ip, port)
         return reader, writer
 
     @async_enterable
-    async def get_stream(self, *command_args, conn_type="I", offset=0):
+    async def get_stream(
+        self,
+        command: Union[str, None] = None,
+        expected_codes: Union[str, tuple[str, ...]] = (),
+        wait_codes: Union[str, tuple[str, ...]] = (),
+        censor_after: Union[int, None] = None,
+        conn_type: ConnType = "I",
+        offset: int = 0,
+    ) -> DataConnectionThrottleStreamIO:
         """
         :py:func:`asyncio.coroutine`
 
@@ -1226,7 +1396,7 @@ class Client(BaseClient):
         reader, writer = await self.get_passive_connection(conn_type)
         if offset:
             await self.command("REST " + str(offset), "350")
-        await self.command(*command_args)
+        await self.command(command, expected_codes, wait_codes, censor_after)
         stream = DataConnectionThrottleStreamIO(
             self,
             reader,
@@ -1236,7 +1406,7 @@ class Client(BaseClient):
         )
         return stream
 
-    async def abort(self, *, wait=True):
+    async def abort(self, *, wait: bool = True) -> None:
         """
         :py:func:`asyncio.coroutine`
 
@@ -1254,14 +1424,14 @@ class Client(BaseClient):
     @contextlib.asynccontextmanager
     async def context(
         cls,
-        host,
-        port=DEFAULT_PORT,
-        user=DEFAULT_USER,
-        password=DEFAULT_PASSWORD,
-        account=DEFAULT_ACCOUNT,
-        upgrade_to_tls=False,
-        **kwargs,
-    ):
+        host: str,
+        port: int = DEFAULT_PORT,
+        user: str = DEFAULT_USER,
+        password: str = DEFAULT_PASSWORD,
+        account: str = DEFAULT_ACCOUNT,
+        upgrade_to_tls: bool = False,
+        **kwargs: Unpack[BaseClientArgs],
+    ) -> AsyncGenerator[Self]:
         """
         Classmethod async context manager. This create
         :py:class:`aioftp.Client`, make async call to
